/*#######################################################################################

Copyright (C) 2013 Leo-Andres Hofmann
www.leo-andres.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*/

#include <avr/io.h>

#include "adc.h"
#include "delay.h"

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Variablen

#if ADC_RUNTIME_CALIBRATION == TRUE

//Zur Laufzeit ermittelte, tatschliche Referenzspannung in mV
static uint16_t adc_calibratedReference = ADC_AREF_MV;

#endif


////////////////////////////////////////////////////////////////////////////////////////////
//Funktionen

//AD-Wandler initialisieren
void adc_init(void)
{
	//Referenzspannung whlen
	ADMUX = ADC_REFERENCE;

	//Vorteiler whlen und AD-Wandler aktivieren
	ADCSRA = ADC_PRESCALER;
	ADCSRA |= (1<<ADEN);

	//Nach dem Aktivieren eine Dummy-Messung auslsen und das Ergebnis verwerfen
	//Gemessen wird die interne Referenz
	adc_read(ADC_MUX_REFERENCE);
}

/******************************************************************************************/

//Einen AD-Kanal lesen
uint16_t adc_read(uint8_t channel)
{
	//Wandler-Kanal whlen, unerlaubte Kanle maskieren und Einstellungen sichern
	ADMUX = (ADMUX & ~(0b1111)) | (channel & 0b1111);

	//Wandlung starten und Ende abwarten
	ADCSRA |= (1<<ADSC);
	while ( ADCSRA & (1<<ADSC) )
	{
	}

	//Messwert zurckgeben
	return ADCW;
}

/******************************************************************************************/

//Einen AD-Kanal lesen und einen Mittelwert bilden
uint16_t adc_readAverage(uint8_t channel)
{
	uint16_t sum = 0;
	uint8_t cycles;

	//Wandler-Kanal whlen, unerlaubte Kanle maskieren und Einstellungen sichern
	ADMUX = (ADMUX & ~(0b1111)) | (channel & 0b1111);

	//Messungen durchfhren
	for(cycles = 0; cycles < ADC_SAMPLES; cycles++)
	{
		//Wandlung starten und Ende abwarten
		ADCSRA |= (1<<ADSC);
		while ( ADCSRA & (1<<ADSC) )
		{
		}

		//Messwerte aufsummieren
		sum += ADCW;
	}

	//Mittelwert bilden und zurckgeben
	return (uint16_t) (sum / ADC_SAMPLES);
}

/******************************************************************************************/

//Gibt einen Spannungswert in mV zurck, Mittelwert-Messung
//Die Spannung wird mit der laufzeitkalibrierten bzw. definierten Referenzspannung ermittelt
uint16_t adc_readVoltage(uint8_t channel)
{
	//Gemessene Spannung berechnen
	//(ADC-Messwert * Referenzspannung in mV) / 1024

#if ADC_RUNTIME_CALIBRATION == TRUE

	return ((uint32_t) (adc_readAverage(channel) * (uint32_t) adc_calibratedReference)) / 1024;

#else

	return ((uint32_t) (adc_readAverage(channel) * (uint32_t) ADC_AREF_MV)) / 1024;

#endif
}

/******************************************************************************************/

#if ADC_RUNTIME_CALIBRATION == TRUE

//Laufzeitkalibrierung auslsen
void adc_runtimeCalibrate(void)
{
	//Dummy-Readout fr maximale Genauigkeit
	adc_read(ADC_MUX_REFERENCE);
	delay_ms(10);

	//Tatschliche Referenzspannung bestimmen
	//(Referenzspannung (mV) * 1024) / ADC-Messwert Referenzspannung
	adc_calibratedReference = ((uint16_t) ADC_AREF_MV * 1024) / adc_readAverage(ADC_MUX_REFERENCE);
}

/******************************************************************************************/

//Gibt die laufzeitkalibrierte Referenzspannung in mV zurck
inline uint16_t adc_getRuntimeReference(void)
{
	return adc_calibratedReference;
}

#endif
