/*#######################################################################################

Copyright (C) 2013 Leo-Andres Hofmann
www.leo-andres.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>

#include "General.h"

////////////////////////////////////////////////////////////////////////////////////////////
//Fehlerprfung der Event-Anzahl

#if GENERAL_EVENTS_COUNT > 8
	#error >8 events configured!
#endif

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Variablen

//Software-Teiler fr das Event-System
static volatile uint8_t general_eventsPrescaler[GENERAL_EVENTS_COUNT] = {};

//Speicher fr Trigger & Status
volatile uint8_t general_eventsTrigger = 0;
struct general_status general_status;

////////////////////////////////////////////////////////////////////////////////////////////
//Funktionen

//Initialisiert den Timer fr das Event-System
void general_initEvents(void)
{
	//Status-Variabel und Software-Teiler zurcksetzen
	general_eventsTrigger = 0;

	for(uint8_t prescaler=0; prescaler<GENERAL_EVENTS_COUNT; prescaler++)
		general_eventsPrescaler[prescaler] = 0;

	//Overflow-Interrupt aktivieren, Prescaler einstellen
	GENERAL_EVENTS_ENABLE;
	GENERAL_EVENTS_CONTROL |= GENERAL_EVENTS_PRESCALER;
}

/******************************************************************************************/

//Watchdog initialisieren
void general_initWatchdog(void)
{
	//Watchdog starten und erstmalig zurcksetzen
	wdt_enable(WATCHDOG_TIMEOUT);
	wdt_reset();
}

/******************************************************************************************/

//Watchdog aktivieren
extern void general_enableWatchdog(void)
{
	wdt_enable(WATCHDOG_TIMEOUT);
}

/******************************************************************************************/

//Watchdog deaktivieren
extern void general_disableWatchdog(void)
{
	wdt_disable();
}

/******************************************************************************************/

//Reset-Status des Controllers auslesen
uint8_t general_readResetState(void)
{
	//Zwischenspeicher fr Rckgabe-Wert
	uint8_t ResetState = 0;

	if ( MCUSR & (1<<BORF) )		// Brown-Out Reset
	{
		ResetState = GENERAL_RESET_BROWN_OUT;
	}
	else if ( MCUSR & (1<<PORF) )	// Power-On Reset
	{
		ResetState = GENERAL_RESET_POWER_ON;
	}
	else if ( MCUSR & (1<<EXTRF) )	// Externer Reset
	{
		ResetState = GENERAL_RESET_EXTERN;
	}
	else if ( MCUSR & (1<<WDRF) )	// Watchdog Reset
	{
		ResetState = GENERAL_RESET_WATCHDOG;
	}
	else							//Kein Ergebnis
	{
		ResetState = GENERAL_RESET_UNKNOWN;
	}

	//Reset-Status rcksetzen & Statuscode zurckgeben
	MCUSR &= 0xF0;

	return ResetState;
}

/******************************************************************************************/

//Statusregister auf Startwerte setzen
void general_initStatus(void)
{
	//Watchdog-Reset aktivieren
	general_status.resetWDT = TRUE;

	general_status.uartSW_State = 0;
	general_status.uartSW_Signal = FALSE;

	general_status.resetSW_State = 0;
	general_status.resetSW_Signal = FALSE;
	general_status.resetSW_Active = FALSE;

	general_status.VCC_USB = OFF;
	general_status.VCC_USB_Overload = FALSE;

	general_status.VCC_MCU = OFF;

	general_status.UART_Link = OFF;
}

/******************************************************************************************/

//Mikrocontroller per Watchdog neu starten
void general_reboot(void)
{
	wdt_enable(WDTO_15MS);

	while(1)
		nop();
}

////////////////////////////////////////////////////////////////////////////////////////////
//Interrupt

//Event-Timer Overflow
ISR(GENERAL_EVENTS_INT)
{
	//Array mit Software-Teilern durchgehen
	for(uint8_t prescaler=0; prescaler<GENERAL_EVENTS_COUNT; prescaler++)
	{
		//Teiler weiterzhlen
		general_eventsPrescaler[prescaler]++;

		//Ermitteln ob ein Teiler den vorgegebenen Wert erreicht hat
		if ( general_eventsPrescaler[prescaler] >= pgm_read_byte_near(&general_eventsSettings[prescaler]) )
		{
			//Teiler zurcksetzen, Event-Flag im Statusregister setzen
			general_eventsPrescaler[prescaler] = 0;
			general_eventsTrigger |= (1<<prescaler);
		}
	}
}
