/*#######################################################################################

miniSTK, PCB-Revision 1.0

Firmware fr ATMega88, Funktionssteuerung, PC-Kommunikation

Einige hardware-spezifische Einstellungen (z.B. fr den Taktausgang, Pin-Change ISR)
werden (noch) fest in der main.c initialisiert und sind nicht einstellbar.

LED Status:
rot							Zielsystem deaktiviert
grn						Zielsystem aktiv, Leerlauf
orange						Zielsystem aktiv, am PC angemeldt
schnell blinken orange		Zugriff auf ISP-Schnittstelle
langsam blinken grn		Verbindung zum PC hergestellt

LED Fehler:
aus							kein Fehler
an							Spannung auerhalb dem erlaubten Bereich
schnell blinken				Fehler im Programmiersystem, PC-Verbindung abgebrochen
langsam blinken				berlastung USB-Anschluss, Kurzschluss

Copyright (C) 2013 Leo-Andres Hofmann
www.leo-andres.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <avr/fuse.h>
#include <avr/lock.h>
#include <avr/signature.h>

#include "settings.h"
#include "general.h"
#include "ports.h"
#include "adc.h"
#include "stk500.h"

////////////////////////////////////////////////////////////////////////////////////////////
//Fuses und Lockbits

FUSES = 
{
	.low		= 0xEF,
	.high		= 0xDD,
	.extended	= 0xFC	//0xFD ohne Bootloader
};

LOCKBITS = ( 0xEE );	//Bootloader & ISP-Flash schreibgeschtzt

////////////////////////////////////////////////////////////////////////////////////////////
//Hardwarespezifische und sonstige Einstellungen
//(nur fr das Hauptprogramm, siehe "settings.h")

//Timer-Ausgang starten/stoppen
#define MAIN_CLK_START				{TCCR1A |= (1<<COM1A0); Clock_On;}
#define MAIN_CLK_STOP				{TCCR1A &= ~(1<<COM1A0); Clock_Off;}

//DEBUG-Modus: Watchdog deaktiviert
//	#define DEBUG

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Variablen

static volatile uint8_t resetIdleCountdown = 0;

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Funktionen

//Taktgeber konfigurieren
//Achtung: Hardwarespezifisch!
static void updateClockTimer(void)
{
	//Taktausgang abschalten
	MAIN_CLK_STOP;

	//Betriebsart: CTC, Toggle OCR1A, TOP = OCR1A
	TCCR1B = (1<<WGM12);

	//Register abhngig vom Vorteiler setzen, verschiedene Korrekturfaktoren beachten:
	//Abweichende Teiler auf neuem Chip, doppelter CPU-Takt
	switch( stk_getClockPrescaler() )
	{
		default:
		case 0: //Teiler 0: Takt gestoppt
			break;

		case 1: //Teiler 1
			TCCR1B |= (1<<CS10); //gesetzt: Teiler 1
			OCR1A = stk_getClockMatch() * 2;
			break;

		case 2: //Teiler 8
			TCCR1B |= (1<<CS11); //gesetzt: Teiler 8
			OCR1A = stk_getClockMatch() * 2;
			break;

		case 3: //Teiler 32
			TCCR1B |= (1<<CS11) | (1<<CS10); //gesetzt: Teiler 64
			OCR1A = stk_getClockMatch();
			break;

		case 4: //Teiler 64
			TCCR1B |= (1<<CS11) | (1<<CS10); //gesetzt: Teiler 64
			OCR1A = stk_getClockMatch() * 2;
			break;

		case 5: //Teiler 128
			TCCR1B |= (1<<CS12); //gesetzt: Teiler 256
			OCR1A = stk_getClockMatch();
			break;

		case 6: //Teiler 256
			TCCR1B |= (1<<CS12); //gesetzt: Teiler 256
			OCR1A = stk_getClockMatch() * 2;
			break;

		case 7: //Teiler 1024
			TCCR1B |= (1<<CS12) | (1<<CS10); //gesetzt: Teiler 1024
			OCR1A = stk_getClockMatch() * 2;
			break;
	}

	//Mindestwert fr Zhler
	if( (stk_getClockPrescaler() > 0) && (OCR1A == 0) )
		OCR1A = 1;
}

////////////////////////////////////////////////////////////////////////////////////////////
int main(void)
{

	uint8_t resetState = 0; //Reset-Status des Controllers (Watchdog, Power-On, ...)
	uint16_t VCC_USB = 0, VCC_MCU = 0; //Gemessene Eingangsspannungen in mV
	uint8_t uartLEDCountdown = 0, stkLEDCountdown = 0, stkSignOnCountdown = 0; //Verzgerung LED-Anzeige

/*----------------------------------------------------------------------------------------*/
	//Init-Sequenz Controller-Status, IO-Pins

	//Reset-Status auslesen
	resetState = general_readResetState();

	//Interrupts lschen
	cli ();

	//Ports initialisieren, alle Funktionen deaktivieren
	ports_init();

	//Reset-Status auswerten, wenn ein Watchdog-Reset ausgelst wurde entsprechenden Status setzen
	if ( resetState == GENERAL_RESET_WATCHDOG )
		general_status.resetByWDT = TRUE;
	else
		general_status.resetByWDT = FALSE;

/*----------------------------------------------------------------------------------------*/
	//Init-Sequenz externe Hardware, Timer & Schnittstellen

#ifndef DEBUG
	//Watchdog-Timer starten mit Einstellungen aus General.h
	general_initWatchdog();
#else
	general_disableWatchdog();
#endif

	//ADC fr Spannungsmessungen
	adc_init();

	//Event-System
	general_initEvents();

	//Pin-Change Interrupt zur Erkennung von Bus-Fehlermeldungen aktivieren (berstrom)
	//Achtung: Hardwarespezifisch!
	PCMSK1 = (1<<PCINT11);
	PCICR |= (1<<PCIE1);

	//Int 0 Interrupt fr USB DTR-Signal
	//Fallende Flanke lst den Interrupt aus
	EICRA = (1<<ISC01);
	EIMSK = (1<<INT0);
	
	//Taktausgang (Timer)
	MAIN_CLK_STOP;

/*----------------------------------------------------------------------------------------*/
	//Init-Sequenz Software

	//Statusregister auf Startwerte setzen	
	general_initStatus();

	//Programmiersystem starten, UART-Verbindung initialisieren
	stk_init();

	//Interrupts einschalten
	sei();

/*----------------------------------------------------------------------------------------*/
	//Hauptschleife

	while(1)
	{
	
		//STK500v2 Protokollverarbeitung, hchste Prioritt
		stk_cronjob();

/*----------------------------------------------------------------------------------------*/
	
		//Event 1: Taster und Statussignale einlesen/entprellen, Reset-Impuls
		if(general_eventTriggered(GENERAL_EVENTS_TRIGGER_1))
		{
			//Trigger zurcksetzen
			general_resetTrigger(GENERAL_EVENTS_TRIGGER_1);


			//Verbindungsaufbau erkennen, LED-Countdown starten
			if( stk_status.signOnOffTrigger == TRUE )
			{
				stk_status.signOnOffTrigger = FALSE;
				stkSignOnCountdown = STK_LED_SIGNON_COUNTDOWN;				
			}

			//Status-LED:	grn blinken nach Verbindungsaufbau
			//				orange blinken bei Portzugriff
			if( stkSignOnCountdown > 0 )
			{
				if(general_status.ledToggle)
					LED_Status_Green
				else
					LED_Status_Off;
			}
			else if( (general_status.resetSW_Active == TRUE) || (stk_status.connectionAssigned == TRUE) || (resetIdleCountdown > (RESET_PULSE_IDLE - STK_LED_COUNTDOWN)) )
			{
				if(general_status.ledFastToggle)
					LED_Status_Orange
				else
					LED_Status_Off;
			}

			//Fehler-LED:	schnell blinken bei Programmierfehler
			if( stk_status.criticalFailure == TRUE )
			{
				if(general_status.ledFastToggle)
					LED_Fault_On
				else
					LED_Fault_Off;
			}

			//Taster entprellen: UART
			if( (general_status.uartSW_State == 0) && (!(SW_UART_DATA & (1<<SW_UART_PIN))) )
			{
				general_status.uartSW_State = 1;
				general_status.uartSW_Signal = TRUE;
			}
			else if( (general_status.uartSW_State == 1) && (!(SW_UART_DATA & (1<<SW_UART_PIN))) )
				general_status.uartSW_State = 2;
			else if( (general_status.uartSW_State == 2) && (SW_UART_DATA & (1<<SW_UART_PIN)) )
				general_status.uartSW_State = 3;
			else if( (general_status.uartSW_State == 3) && (SW_UART_DATA & (1<<SW_UART_PIN)) )
				general_status.uartSW_State = 0;

			//Taster entprellen: Reset
			if( (general_status.resetSW_State == 0) && (!(SW_RESET_DATA & (1<<SW_RESET_PIN))) )
			{
				general_status.resetSW_State = 1;
				general_status.resetSW_Signal = TRUE;
			}
			else if( (general_status.resetSW_State == 1) && (!(SW_RESET_DATA & (1<<SW_RESET_PIN))) )
				general_status.resetSW_State = 2;
			else if( (general_status.resetSW_State == 2) && (SW_RESET_DATA & (1<<SW_RESET_PIN)) )
				general_status.resetSW_State = 3;
			else if( (general_status.resetSW_State == 3) && (SW_RESET_DATA & (1<<SW_RESET_PIN)) )
				general_status.resetSW_State = 0;

			//Countdown Verzgerungen
			if( uartLEDCountdown > 0 )
				uartLEDCountdown--;
			if( stkLEDCountdown > 0 )
				stkLEDCountdown--;
			if( stkSignOnCountdown > 0 )
				stkSignOnCountdown--;
			if( resetIdleCountdown > 0 )
				resetIdleCountdown--;

			//Toggle-Bit fr LED-Anzeige umschalten
			general_status.ledFastToggle ^= 1;
		}

/*----------------------------------------------------------------------------------------*/
	
		//Event 2: Spannungsmessung, Statusanzeige, Watchdog-Reset, UART-Schaltung
		if(general_eventTriggered(GENERAL_EVENTS_TRIGGER_2))
		{
			//Trigger zurcksetzen
			general_resetTrigger(GENERAL_EVENTS_TRIGGER_2);

			//Spannungsmessung
			VCC_USB = adc_readVoltage(ADC_VCCIO_PIN);
			VCC_MCU = adc_readVoltage(ADC_VCC_PIN);

			//Auswertung nach Schwellwerten
			if( VCC_USB >= VCC_OFF )
			{
				if( (VCC_USB >= VCC_MIN) && (VCC_USB <= VCC_MAX) )
					general_status.VCC_USB = ON;
				else
					general_status.VCC_USB = BAD;	
			}
			else
			{
				general_status.VCC_USB = OFF;
			}

			if( VCC_MCU >= VCC_OFF )
			{
				if( (VCC_MCU >= VCC_MIN) && (VCC_MCU <= VCC_MAX) )
					general_status.VCC_MCU = ON;
				else
					general_status.VCC_MCU = BAD;	
			}
			else
			{
				general_status.VCC_MCU = OFF;
			}

			//Gemessene Spannung an STK-Protokoll bergeben
			//Umrechung in 1/10 Volt, Spannungsteiler 5.7:1
			stk_assignVTarget((VCC_MCU * 57) / 1000);

			//Taktausgang: Neue Einstellungen bernehmen
			if( stk_status.updateClock == TRUE )
			{
				stk_status.updateClock = FALSE;
				updateClockTimer();
			}

			//Taktausgang: Deaktivierem bei ausgeschalteter Versorgung
			if( general_status.VCC_MCU != ON)
				MAIN_CLK_STOP
			else if( stk_getClockPrescaler() > 0 )
				MAIN_CLK_START;

			//berlastmeldung lschen wenn die USB-Versorgung unterbrochen ist
			if( (general_status.VCC_USB == OFF) && (USB_OVERLOAD_LATCHING == FALSE) )
				general_status.VCC_USB_Overload = FALSE;


			//Fehler-LED: 	Blinken bei berlastung
			//				Dauerleuchten bei VCC auer Bereich
			//nicht umschalten bei Fehlermeldung STK
			if( stk_status.criticalFailure == FALSE )
			{
				if( general_status.VCC_USB_Overload == TRUE )
				{
					if( general_status.ledToggle )
						LED_Fault_On
					else
						LED_Fault_Off;
				}
				else if( (general_status.VCC_USB == BAD) || (general_status.VCC_MCU == BAD) )
				{
					LED_Fault_On;
				}
				else
				{
					LED_Fault_Off;
				}
			}

			//Status-LED:	grn bei aktivem Zielsystem, Leerlauf
			//				orange bei Verbindung zum Programmiersystem
			//				rot bei deaktiviertem Zielsystem
			//nicht umschalten wenn ein Countdown/Fehlermeldung aktiv ist
			if( (stkSignOnCountdown <= 0) && (general_status.resetSW_Active == FALSE) && (stk_status.connectionAssigned == FALSE) )
			{
				if( general_status.VCC_MCU != ON )
				{
					LED_Status_Red;
				}
				else if( stk_status.signedOn == TRUE )
				{
					LED_Status_Orange;
				}
				else if( general_status.resetSW_Active == FALSE )
				{
					LED_Status_Green;
				}
			}

			//Unzulssige Verbindungen verhindern
			if( general_status.VCC_USB != ON ) //keine USB-Verbindung
				general_status.UART_Link = OFF;
			else if( (general_status.VCC_USB == ON) && (general_status.VCC_MCU != ON) ) //Zielsystem aus
				general_status.UART_Link = STK;
			else if( general_status.UART_Link == OFF ) //USB Verbindung steht und Zielsystem ist an, Verbindung war getrennt
#if UART_START_STK == TRUE
				general_status.UART_Link = STK;
#else
				general_status.UART_Link = MCU;
#endif

			//UART-LED & Treiber umschalten
			if( general_status.UART_Link == OFF )
			{
				UART_Disable;
				UART_Target_STK;

				LED_UART_Off;
			}
			else if( (general_status.UART_Link == MCU) && (uartLEDCountdown <= 0) )
			{
				UART_Enable;
				UART_Target_MCU;

				LED_UART_MCU;
			}
			else if( (general_status.UART_Link == STK) && (uartLEDCountdown <= 0) )
			{
				UART_Enable;
				UART_Target_STK;

				LED_UART_STK;
			}
		
			//Toggle-Bit fr LED-Anzeige umschalten
			general_status.ledToggle ^= 1;

			//WDT zurcksetzen, sofern noch aktiviert (wird zum selbst-Reset abgeschaltet)
			if ( general_status.resetWDT == TRUE )
			     wdt_reset();
		}

/*----------------------------------------------------------------------------------------*/
	
		//Event 3: Statusanzeige, Taster auswerten
		if(general_eventTriggered(GENERAL_EVENTS_TRIGGER_3))
		{
			//Trigger zurcksetzen
			general_resetTrigger(GENERAL_EVENTS_TRIGGER_3);

			//STK-Verbindung trennen/Fehler zurcksetzen wenn beide Taster bettigt sind
			if( (general_status.uartSW_Signal == TRUE) && (general_status.resetSW_Signal == TRUE) && (stk_status.signedOn == TRUE) )
			{
				//Trigger zurcksetzen
				general_status.uartSW_Signal = FALSE;
				general_status.resetSW_Signal = FALSE;

				stk_disconnect();
			}

			//Taster bettigt: UART
			if( general_status.uartSW_Signal == TRUE )
			{
				//Trigger zurcksetzen
				general_status.uartSW_Signal = FALSE;

				//UART-Verbindung umschalten
				//nur bei zwei mglichen Verbindungen umschalten, gesperrt bei Aktivitt
				if( (general_status.VCC_USB == ON) && (general_status.VCC_MCU == ON) && (stk_status.signedOn == FALSE) )
				{
					if( general_status.UART_Link == STK )
						general_status.UART_Link = MCU;
					else
						general_status.UART_Link = STK;

					//Verbindung wird in Event 2 passend neu aufgebaut
					UART_Disable;

					//Beide UART-LEDs einschalten und Verzgerung starten
					uartLEDCountdown = UART_LED_COUNTDOOWN;
					LED_UART_Both;
				}
			}

			//Taster bettigt: Reset
			if( general_status.resetSW_Signal == TRUE )
			{
				//Trigger zurcksetzen
				general_status.resetSW_Signal = FALSE;

				//Reset-Impuls nicht auslsen wenn das Programmiersystem zugreift oder die Interrupt-Sperrzeit noch abluft
				if( (general_status.VCC_MCU == ON) && (stk_status.connectionAssigned == FALSE) && (resetIdleCountdown == 0) )
				{
					stkT_ISP_tgStop();

					LED_Status_Orange;
					stkLEDCountdown = STK_LED_COUNTDOWN;
					general_status.resetSW_Active = TRUE;
				}					
			}

			//Aktiven Reset-Impuls zurcknehmen wenn der Countdown abgelaufen ist
			//und der Taster nicht mehr gehalten wird (Impulsverlngerung)
			if( (general_status.resetSW_Active == TRUE) && (stkLEDCountdown == 0) && (SW_RESET_DATA & (1<<SW_RESET_PIN)) && (stk_status.connectionAssigned == FALSE) )
			{
				stkT_ISP_tgRun();

				general_status.resetSW_Active = FALSE;
			}
		}
	}

	return 0;
}

////////////////////////////////////////////////////////////////////////////////////////////
//Interrupts

//Pin-Change Interrupt zur schnellen Erkennung von Bus-Fehlermeldungen (berstrom)
ISR(PCINT1_vect)
{
	//berstrom-Meldung erkennen wenn der Pin als Eingang geschaltet ist
	if( (!(USB_FAULT_DATA & (1<<USB_FAULT_PIN))) && (USB_Fault_isInput) )
	{
		//UART-Verbindung deaktivieren
		UART_Disable;
		general_status.UART_Link = OFF;

		//berstrom-Meldung setzen
		general_status.VCC_USB_Overload = TRUE;
	}	
	else if( USB_OVERLOAD_LATCHING == FALSE )
	{
		//berstrom-Meldung lschen, wenn eine steigende Flanke erkannt wurde
		general_status.VCC_USB_Overload = FALSE;
	}
}

/******************************************************************************************/

//Interrupt 0: USB DTR-Signal zum ferngesteuerten Reset (fallende Flanke)
ISR(INT0_vect)
{
	if( (general_status.UART_Link == MCU) && (stk_status.connectionAssigned == FALSE) && (resetIdleCountdown == 0) )
	{
		stkT_ISP_tgResetPulse();
		
		LED_Status_Orange;
		resetIdleCountdown = RESET_PULSE_IDLE;		
	}
}
