/*#######################################################################################

Copyright (C) 2013 Leo-Andres Hofmann
www.leo-andres.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*/

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <avr/wdt.h>

#include "stk500.h"
#include "stk500_tools.h"
#include "stk500_protocol.h"

#include "uart.h"
#include "delay.h"

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Variablen

static uint8_t stk_targetVoltage = 0;
static const char stk_signature[8] PROGMEM = STK_CONF_SIGNATURE;

static uint8_t stk_sequence = 0; //Zwischenspeicher fr Sequenznummern

//Adressen
static uint32_t stk_address = 0; //Speicheradresse
static uint8_t stk_extendedAddress = 0; //Erweiterte Adresse, Bit 16-24

//Taktgeber-Einstellung
static uint8_t stk_clockPrescaler, stk_clockMatch;

//Speicher fr Status
struct stk_status stk_status;
volatile struct stk_frame stk_frame;

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Funktionen

//Antwort-Code setzen, auerdem setzt diese Funktion die Lnge auf 2 Bytes
static void stk_setResponse(uint8_t status);

//Parameter-Anwort setzen, Lnge 3
static void stk_setParamResponse(uint8_t param, uint8_t status);

////////////////////////////////////////////////////////////////////////////////////////////
//Funktionen

//Initialisiert die UART-Schnittstelle und die Protokollverarbeitung
void stk_init(void)
{
	//Statusregister initialisieren
	stk_status.signedOn = FALSE;
	stk_status.signOnOffTrigger = FALSE;
	stk_status.connectionAssigned = FALSE;
	stk_status.criticalFailure = FALSE;
	stk_status.updateClock = FALSE;

	stk_status.extendedAddressing = FALSE;
	stk_status.reloadAddress = FALSE;
	stk_status.byteAddressing = FALSE;
	stk_status.disablePolling = FALSE;

	stk_sequence = 0;
	stk_address = 0;

	stk_targetVoltage = 0;

	stk_clockPrescaler = STK_CLOCK_DEFAULT_PRESCALER;
	stk_clockMatch = STK_CLOCK_DEFAULT_MATCH;

#if STK_CLOCK_DEFAULT_PRESCALER > 0
	stk_status.updateClock = TRUE;
#endif

	//UART-Verbindung aufbauen
	uart_init();
	
	//Port freigeben, Verbindung ist getrennt
	stkT_ISP_release();	
}

/******************************************************************************************/

//Cronjob (Frame-Verarbeitung)
//Diese Funktion muss periodisch aufgerufen werden
//Achtung: Nicht in einem Interrupt aufrufen!
void stk_cronjob(void)
{	
	static uint8_t controlByte = 0; //berwachungsbyte Verbindungsverlust

	uint8_t bytePos, ispData, ispSync, receiveStart, progMode;
	uint16_t transferCnt, transferPos, pollAddress, pageAddress;

	//Adressierung: Word, nur bei EEPROM-Zugriffen Byte
	stk_status.byteAddressing = FALSE;

	//Nur vollstndige Frames verarbeiten
	if( uart_status.frameComplete == TRUE )
	{
		//Trigger zurcksetzen
		uart_status.frameComplete = FALSE;

		stk_sequence++;

		//Sequenznummernprfung: Die empfangene Sequenznummer ist immer +1 zur vorigen
		//Ausnahme: Beim Verbindungsaufbau werden beide Zhler synchronisiert
		if( (stk_frame.sequence != stk_sequence) && (stk_frame.data[0] != STK_CMD_SIGN_ON) )
			uart_status.frameError = TRUE;

		if( uart_status.frameError == TRUE )
		{
			//Frame-Fehler: Fehler-Status senden
			stk_setResponse(STK_STAT_CMD_FAILED);

			//Sonderfall Prfsummen-Fehler: Kommando-ID ersetzen
			if( uart_status.frameChecksumError == TRUE )
			{
				stk_frame.data[0] = STK_ANSW_CHECKSUM_ERROR;
				stk_setResponse(STK_STAT_CMD_CHECKSUM_ERROR);
			}

			if( stk_status.signedOn == TRUE )
				stk_status.criticalFailure = TRUE;
		}
		else
		{
			//Eingehende Kommandos verarbeiten
			switch( stk_frame.data[0] )
			{
				default:
					stk_setResponse(STK_STAT_CMD_UNKNOWN);
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_SIGN_ON:
					stk_frame.dataLength = 11;
					stk_frame.data[1] = STK_STAT_CMD_OK;
					stk_frame.data[2] = 8;

					//Signatur anhngen
					for(bytePos=0; bytePos<8; bytePos++)
						stk_frame.data[bytePos + 3] = pgm_read_byte_near(&stk_signature[bytePos]);

					//Sonderfall Verbindungsaufbau: Sequenznummer synchronisieren, Status zurcksetzen
					stk_sequence = stk_frame.sequence;
					stk_status.criticalFailure = FALSE;

					stk_status.signedOn = TRUE;
					stk_status.signOnOffTrigger = TRUE;
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_SET_PARAM:
					//1: Parameter
					//2: Wert

					switch( stk_frame.data[1] )
					{
						default:
						case STK_PARAM_VTARGET:		//nicht untersttzt
						case STK_PARAM_VADJUST:						
							stk_setResponse(STK_STAT_CMD_UNKNOWN);
							break;

						case STK_PARAM_OSC_PSCALE:
							stk_clockPrescaler = stk_frame.data[2];
							stk_status.updateClock = TRUE;

							stk_setResponse(STK_STAT_CMD_OK);
							break;

						case STK_PARAM_OSC_CMATCH:							
							stk_clockMatch = stk_frame.data[2];
							stk_status.updateClock = TRUE;

							stk_setResponse(STK_STAT_CMD_OK);
							break;

						case STK_PARAM_SCK_DURATION:
							stkT_ISP_setClkDuration(stk_frame.data[2]);
							stk_setResponse(STK_STAT_CMD_OK);
							break;

						case STK_PARAM_CONTROLLER_INIT:
							controlByte = stk_frame.data[2];
							stk_setResponse(STK_STAT_CMD_OK);
							break;

						case STK_PARAM_RESET_POLARITY: //Reset-Polaritt kann nicht verndert werden, Parameter muss fr den Verbindungsaufbau trotzdem akzeptiert werden
							stk_setResponse(STK_STAT_CMD_OK);
							break;
					}
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_GET_PARAM:
					//1: Parameter

					switch( stk_frame.data[1] )
					{
						default:
							stk_setResponse(STK_STAT_CMD_UNKNOWN);
							break;

						case STK_PARAM_BUILD_NUMBER_LOW:
							stk_setParamResponse(STK_CONF_PARAM_BUILD_NUMBER_LOW, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_BUILD_NUMBER_HIGH:
							stk_setParamResponse(STK_CONF_PARAM_BUILD_NUMBER_HIGH, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_HW_VER:
							stk_setParamResponse(STK_CONF_PARAM_HW_VER, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_SW_MAJOR:
							stk_setParamResponse(STK_CONF_PARAM_SW_MAJOR, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_SW_MINOR:
							stk_setParamResponse(STK_CONF_PARAM_SW_MINOR, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_VTARGET:
							stk_setParamResponse(stk_targetVoltage, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_VADJUST: //keine Spannung
							stk_setParamResponse(0, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_OSC_PSCALE: //Teiler 0: Takt deaktiviert wenn Spannung ausgeschaltet
							if( general_status.VCC_MCU == ON )
								stk_setParamResponse(stk_clockPrescaler, STK_STAT_CMD_OK);
							else
								stk_setParamResponse(0, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_OSC_CMATCH:
							stk_setParamResponse(stk_clockMatch, STK_STAT_CMD_OK);
							break;

						case STK_PARAM_SCK_DURATION:
							stk_setParamResponse(stkT_ISP_getClkDuration(), STK_STAT_CMD_OK);
							break;

						case STK_PARAM_TOPCARD_DETECT:
							stk_setParamResponse(STK_CONF_PARAM_TOPCARD, STK_STAT_CMD_OK);
							break;

						/*case STK_PARAM_STATUS: //Statusregister
							stk_setParamResponse(0,STK_STAT_CMD_OK);
							break;*/

						/*case STK_PARAM_DATA: //Datenregister fr HV/PP, nicht untersttzt
							stk_setParamResponse(0,STK_STAT_CMD_OK);
							break;*/
						
						case STK_PARAM_CONTROLLER_INIT:
							stk_setParamResponse(controlByte, STK_STAT_CMD_OK);
							break;
					}
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_FW_UPGRADE:
					stk_setResponse(STK_STAT_CMD_FAILED); //Update wird ber dieses Kommando nicht untersttzt
					//general_reboot(); //Neustart in Bootloader
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_LOAD_ADDR:
					//1-4: Adresse, MSB zuerst
					//64k Addressierung bei Adressen > 2MB
					stk_address =  ((uint32_t) stk_frame.data[1]) << 24;
					stk_address |= ((uint32_t) stk_frame.data[2]) << 16;
					stk_address |= ((uint32_t) stk_frame.data[3]) << 8;
					stk_address |= ((uint32_t) stk_frame.data[4]);

					stk_extendedAddress = stk_frame.data[2];

					//Bit 31 gesetzt: 64k Adressierung
					if( stk_frame.data[1] & (1<<7) ) 
						stk_status.extendedAddressing = TRUE;
					else
						stk_status.extendedAddressing = FALSE;

					stk_status.reloadAddress = TRUE;

					stk_setResponse(STK_STAT_CMD_OK);
					break;


// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_ENTER_PROGMODE_ISP:
					//Programmiermodus nicht starten wenn das Zielsystem inaktiv ist
					if( general_status.VCC_MCU != ON )
					{
						stk_setResponse(STK_STAT_CMD_FAILED);
						break;
					}

					//1: Timeout, nicht untersttzt
					//2: Stabilisierungszeit nach Verbindung
					//3: Wartezeit vor bertragung
					//4: Anzahl Verbindungsversuche
					//5: Wartezeit zwischen Bytes
					//6: Synchronisation Testwert
					//7: Synchronisation Prfbyte Index (Testwert muss am Index zurckgelesen werden)
					//8-11: Kommandobytes

					ispSync = FALSE;

					//Mindestwartezeit 1ms
					if( stk_frame.data[5] <= 0 )
						stk_frame.data[5] = 1;
					
					//Schnittstelle starten
					stkT_ISP_assign();
					stkT_ISP_tgStop();

					//Stabilisierung abwarten
					delay_ms(stk_frame.data[2]);

					do
					{

#if STK_WATCHDOG_RESET == TRUE
						wdt_reset();
#endif

						//Wartezeit vor bertragungsstart
						delay_ms(stk_frame.data[3]);

						for(bytePos=0; bytePos<4; bytePos++)
						{
							//Kommanobyte senden und abwarten
							ispData = stkT_ISP_transfer_8(stk_frame.data[bytePos + 8]);
							delay_ms(stk_frame.data[5]);

							//Auf Synchronisation prfen wenn eine Prfbyte-Position gesetzt ist
							if( (stk_frame.data[7] > 0) && (stk_frame.data[7] == bytePos + 1) )
							{
								//Prfbyte zurckgelesen, Synchronsation hergestellt
								if( ispData == stk_frame.data[6] )
									ispSync = TRUE;
							}
							else if( stk_frame.data[7] == 0 ) //Keine Prfung aktiviert, nur einmalig senden und Ende
								ispSync = TRUE;
						}

						if( ispSync == TRUE )
							break;

						//Bus 1 Bit weitertakten bis Synchronisation einraster
						stkT_ISP_clockPulse();

					} while( --stk_frame.data[4] );

					//Synchronisation hergestellt oder kein Prfwert verwendet
					if( (ispSync == TRUE) || (stk_frame.data[7] == 0) )
						stk_setResponse(STK_STAT_CMD_OK);
					else
						stk_setResponse(STK_STAT_CMD_TOUT);
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_LEAVE_PROGMODE_ISP:
					stkT_ISP_release();

					stk_setResponse(STK_STAT_CMD_OK);
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_CHIP_ERASE_ISP:
					//1: Wartezeit nach dem Lschen
					//2: Abfragemethode
					//3-6: Kommando

					ispSync = FALSE;

					//Kommanobytes senden 
					for(bytePos=0; bytePos<4; bytePos++)
						stkT_ISP_transfer_8(stk_frame.data[bytePos + 3]);					

#if STK_WATCHDOG_RESET == TRUE
					wdt_reset();
#endif

					//Status abfragen/abwarten
					if( stk_frame.data[2] == 1 )
					{
						if( stkT_ISP_TestRDYBSY(STK_RDYBSY_CYCLES) )
							ispSync = TRUE;
					}
					else
						delay_ms(stk_frame.data[1]);

					//Lschen beendet oder keine Synchronisation abfragen
					if( (ispSync == TRUE) || (stk_frame.data[2] == 0) )
						stk_setResponse(STK_STAT_CMD_OK);
					else
						stk_setResponse(STK_STAT_CMD_TOUT);
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_PROGRAM_EEPROM_ISP:
					//EEPROM-Zugriff byteweise
					stk_status.byteAddressing = TRUE;

				case STK_CMD_PROGRAM_FLASH_ISP:
					//1-2: Anzahl Bytes, MSB zuerst
					//3: Modus-Byte
					//4: Wartezeit Schreibzugriff
					//5: Kommando Speicherseite laden, Byte schreiben
					//6: Kommando Speicherseite schreiben
					//7: Kommando Byte lesen
					//8: Prfbyte 1, Schreibzugriff Ende prfen
					//9: Prfbyte 2
					//10-n: Daten

					//Modus-Byte: Statusbits
					//0: Word (0) oder Seiten (1) adressieren
					//1: Word: Schreibzugriff abwarten 
					//2: Word: Wertprfung
					//3: Word: RDY/BSY Prfung
					//4: Seite: Schreibzugriff abwarten 
					//5: Seite: Wertprfung
					//6: Seite: RDY/BSY Prfung
					//7: Aktuelle Speicherseite schreiben

					//ISP-Kommando
					//1: Kommando, Bit 3 gesetzt um das High-Byte zu lesen
					//2: Adresse MSB
					//3: Adresse LSB
					//4: Datenbyte

					//Startwerte
					stk_status.disablePolling = FALSE;
					pollAddress = 0;
					pageAddress = stk_address & 0xFFFF;

					ispSync = STK_STAT_CMD_FAILED;

					//Mindestwartezeit 10ms
					if( stk_frame.data[4] <= 0 )
						stk_frame.data[4] = 10;

					transferCnt =  ((uint32_t) stk_frame.data[1]) << 8;
					transferCnt |= ((uint32_t) stk_frame.data[2]);
					progMode = stk_frame.data[3];

					//Maximale Pufferlnge, abzglich Statusbits
					if( transferCnt > STK_MSG_BUFFER - 3)
					{
						stk_setResponse(STK_STAT_CMD_FAILED);
						break;
					}

					//Modus Bit 0: Word (0) oder Seiten (1) adressieren
					if( progMode & (1<<0) )
					{
						//Schreibkommando siehe 5
						ispData = stk_frame.data[5];

						//Speicher schreiben
						for(transferPos=0; transferPos<transferCnt; transferPos++)
						{

#if STK_WATCHDOG_RESET == TRUE
							wdt_reset();
#endif
							//64k Adressierung: ber der 64k Grenze bei jedem Zugriff neu laden, unterhalb nur bei neuen Adressen
							if( (stk_status.extendedAddressing == TRUE) && ((stk_status.reloadAddress == TRUE) || (stk_address > 0xFFFF)) )
							{
								stk_status.reloadAddress = FALSE;

								//Addressierungskommando 0x4D
								stkT_ISP_transfer_8(0x4D);
								stkT_ISP_transfer_8(0x00);
								stkT_ISP_transfer_8(stk_extendedAddress);
								stkT_ISP_transfer_8(0x00);
							}

							//1: Kommandobyte senden: Bei Word-Adressen muss Bit 3 im Kommando gesetzt sein um das High-Byte anzusprechen
							//Das High-Byte wird immer auf ungerade Speicherposition im Puffer abgelegt
							if( (stk_status.byteAddressing == FALSE) && (transferPos % 2) )
								stkT_ISP_transfer_8(ispData | (1<<3));
							else
								stkT_ISP_transfer_8(ispData);

							//2-3: Adresse senden
							stkT_ISP_transfer_16(stk_address & 0xFFFF);

							//4: Byte lesen
							stkT_ISP_transfer_8(stk_frame.data[transferPos + 10]);

							//Das letzte geschrieben Byte setzt jeweils die Polling-Address fr die Schreibzugriff-Prfung
							//(nur mglich wenn das Prfbyte vom Datenbyte abweicht)
							//pollAddress = stk_address & 0xFFFF;
							if( stk_frame.data[transferPos + 10] == stk_frame.data[8] )
								stk_status.disablePolling = TRUE;
							else
								pollAddress = stk_address & 0xFFFF;

							//Adresse weiterzhlen
							//Word-Adressierung, nur bei ungeraden Speicherpositionen (nach dem High-Byte) weiterzhlen
							if( (stk_status.byteAddressing == FALSE) && (transferPos % 2) )
							{
								stk_address++;
								if( stk_address > 0xFFFF )
									stk_extendedAddress++;
							}
							else if( stk_status.byteAddressing == TRUE ) //Byte-Adressierung, weiterzhlen nach jedem Transfer
								stk_address++;
						}

						//Modus Bit 7: Speicherseite schreiben
						if( progMode & (1<<7) )
						{
							//Schreibkommando senden, Kommandobyte siehe 6
							stkT_ISP_transfer_8(stk_frame.data[6]);
							stkT_ISP_transfer_16(pageAddress);
							stkT_ISP_transfer_8(0x00);

#if STK_WATCHDOG_RESET == TRUE
							wdt_reset();
#endif

							//Schreibzugriff abfragen
							if( (progMode & (1<<5)) && (stk_status.disablePolling == FALSE) ) //Wert-Prfung
							{
								//Lesekommando siehe 7
								ispData = stk_frame.data[7];

								bytePos = STK_POLLING_CYCLE;

								while( bytePos > 0) 
								{
									//Kommandobyte, Bit 3 setzen um bei ungeraden Adressen das High-Byte anzusprechen (siehe Transfer)
									if( pollAddress % 2 )
										stkT_ISP_transfer_8(ispData | (1<<3));
									else
										stkT_ISP_transfer_8(ispData);

									stkT_ISP_transfer_16(pollAddress);

									//Der Schreibzugriff ist beendet wenn der Chip mit einem abweichenden Wert antwortet
									//Prfbyte siehe 8
									if( stkT_ISP_transfer_8(0x00) != stk_frame.data[8] )
										break;

									bytePos--;
								}

								if( bytePos )
									ispSync = STK_STAT_CMD_OK;
								else
									ispSync = STK_STAT_CMD_TOUT;
							}
							else if( progMode & (1<<6) ) //RDY/BSY Prfung
							{
								if( stkT_ISP_TestRDYBSY(STK_RDYBSY_CYCLES) )
									ispSync = STK_STAT_CMD_OK;
								else
									ispSync = STK_STAT_RDY_BSY_TOUT;
							}
							else //Wartezeit
							{
								delay_ms(stk_frame.data[4]);
								ispSync = STK_STAT_CMD_OK;
							}

						}
						else //Seite unvollstndig, bertragung besttigen und vollstndige Seite abwarten
							ispSync = STK_STAT_CMD_OK;
					}
					else //Word-Adressierung
					{
						//TODO: Einbauen fr ltere Chips!

						//Speicher schreiben
					/*	for(transferPos=0; transferPos<transferCnt; transferPos++)
						{

#if STK_WATCHDOG_RESET == TRUE
							wdt_reset();
#endif
						}*/
					}

					stk_setResponse(ispSync);
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_READ_EEPROM_ISP:
					//EEPROM-Zugriff byteweise
					stk_status.byteAddressing = TRUE;

				case STK_CMD_READ_FLASH_ISP:
					//1-2: Anzahl Bytes, MSB zuerst
					//3: Lesekommando Byte 1

					//ISP-Kommando:
					//1: Kommando, Bit 3 gesetzt um das High-Byte zu lesen
					//2: Adresse MSB
					//3: Adresse LSB
					//4: 0x00 <- Byte abfragen

					transferCnt =  ((uint32_t) stk_frame.data[1]) << 8;
					transferCnt |= ((uint32_t) stk_frame.data[2]);
					ispData = stk_frame.data[3];

					//Maximale Pufferlnge, abzglich Statusbits
					if( transferCnt > STK_MSG_BUFFER - 3)
					{
						stk_setResponse(STK_STAT_CMD_FAILED);
						break;
					}

					//Speicher auslesen
					for(transferPos=0; transferPos<transferCnt; transferPos++)
					{

#if STK_WATCHDOG_RESET == TRUE
						wdt_reset();
#endif

						//64k Adressierung: ber der 64k Grenze bei jedem Zugriff neu laden, unterhalb nur bei neuen Adressen
						if( (stk_status.extendedAddressing == TRUE) && ((stk_status.reloadAddress == TRUE) || (stk_address > 0xFFFF)) )
						{
							stk_status.reloadAddress = FALSE;

							//Addressierungskommando 0x4D
							stkT_ISP_transfer_8(0x4D);
							stkT_ISP_transfer_8(0x00);
							stkT_ISP_transfer_8(stk_extendedAddress);
							stkT_ISP_transfer_8(0x00);
						}

						//1: Kommandobyte senden: Bei Word-Adressen muss Bit 3 im Kommando gesetzt sein um das High-Byte anzusprechen
						//Das High-Byte wird immer auf ungerade Speicherposition im Puffer abgelegt
						if( (stk_status.byteAddressing == FALSE) && (transferPos % 2) )
							stkT_ISP_transfer_8(ispData | (1<<3));
						else
							stkT_ISP_transfer_8(ispData);
					
						//2-3: Adresse senden
						stkT_ISP_transfer_16(stk_address & 0xFFFF);

						//4: Byte lesen
						stk_frame.data[transferPos + 2] = stkT_ISP_transfer_8(0x00);

						//Adresse weiterzhlen
						//Word-Adressierung, nur bei ungeraden Speicherpositionen (nach dem High-Byte) weiterzhlen
						if( (stk_status.byteAddressing == FALSE) && (transferPos % 2) )
						{
							stk_address++;
							if( stk_address > 0xFFFF )
								stk_extendedAddress++;
						}
						else if( stk_status.byteAddressing == TRUE ) //Byte-Adressierung, weiterzhlen nach jedem Transfer
							stk_address++;
					}

					stk_frame.dataLength = transferCnt + 3;
					stk_frame.data[1] = STK_STAT_CMD_OK;
					stk_frame.data[transferCnt + 2] = STK_STAT_CMD_OK;
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_SPI_MULTI:
					//1: Anzahl Bytes senden
					//2: Anzahl Bytes empfangen
					//3: Start angeforderte Daten
					//4-n: Sendedaten, Anzahl siehe 1

					transferCnt = stk_frame.data[2];
					receiveStart = stk_frame.data[3];

					//Prfen ob mehr Daten gelesen werden mssen
					if( transferCnt + receiveStart > stk_frame.data[1] )
						transferCnt = transferCnt + receiveStart;
					else
						transferCnt = stk_frame.data[1];

					//bertragung
					for(bytePos=0; bytePos<transferCnt; bytePos++)
					{
						//Anzahl Sendebytes aus Puffer bertragen, Rest mit 0x00 fllen
						if(bytePos <  stk_frame.data[1])
							ispData = stkT_ISP_transfer_8(stk_frame.data[bytePos + 4]);
						else
							ispData = stkT_ISP_transfer_8(0x00);

						//Empfagene Daten in Sendepuffer legen ab Startposition
						if( bytePos >= receiveStart )
							stk_frame.data[(bytePos - receiveStart) + 2] = ispData;
					}

					stk_frame.dataLength = transferCnt + 3;
					stk_frame.data[1] = STK_STAT_CMD_OK;
					stk_frame.data[transferCnt + 2] = STK_STAT_CMD_OK;
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_PROGRAM_FUSE_ISP:
				case STK_CMD_PROGRAM_LOCK_ISP:
					//1-4: Kommando

					for(bytePos=0; bytePos<4; bytePos++)
						stkT_ISP_transfer_8(stk_frame.data[bytePos + 1]);

					stk_frame.dataLength = 3;
					stk_frame.data[1] = STK_STAT_CMD_OK;
					stk_frame.data[2] = STK_STAT_CMD_OK;
					break;

// // // // // // // // // // // // // // // // // // // // // // // // // // // // // // //
				case STK_CMD_READ_FUSE_ISP:
				case STK_CMD_READ_LOCK_ISP:
				case STK_CMD_READ_SIGNATURE_ISP:
				case STK_CMD_READ_OSCCAL_ISP:
					//1: Angefordertes Byte
					//2-5: Kommando

					ispData = 0xFF;

					//Kommandos bertragen und Antwort sichern
					for(bytePos=0; bytePos<4; bytePos++)
					{
						if( stk_frame.data[1] == bytePos + 1 )
							ispData = stkT_ISP_transfer_8(stk_frame.data[bytePos + 2]);
						else
							stkT_ISP_transfer_8(stk_frame.data[bytePos + 2]);
					}

					stk_frame.dataLength = 4;
					stk_frame.data[1] = STK_STAT_CMD_OK;
					stk_frame.data[2] = ispData;
					stk_frame.data[3] = STK_STAT_CMD_OK;
					break;
			}
		}

		//Fertig aufgebauten Frame senden
		stk_frame.sequence = stk_sequence;
		uart_sendFrame();

		//Vorbereiten fr den nchsten Frame-Empfang
		uart_resetReceiver();
	}
}

/******************************************************************************************/

//Zielsystem-Spannung zuweisen fr Statusmeldung an den PC
//Einheit: 1/10 Volt (42 = 4.2V)
inline void stk_assignVTarget(uint8_t voltage)
{
	stk_targetVoltage = voltage;

	//Verbindung unterbrechen wenn das Zielsystem inaktiv ist
	if( (general_status.VCC_MCU != ON) && (stk_status.connectionAssigned == TRUE) )
	{
		stkT_ISP_release();

		stk_status.criticalFailure = TRUE;
		stk_sequence = 0;
	}
}

/******************************************************************************************/

//Verbindung trennen, Ports freigeben
void stk_disconnect(void)
{
	if( stk_status.connectionAssigned == TRUE )
		stkT_ISP_release();

	stk_status.signedOn = FALSE;	
	stk_status.signOnOffTrigger = TRUE;

	stk_status.criticalFailure = FALSE;

	stk_sequence = 0;
	stk_address = 0;
}

/******************************************************************************************/

//Taktgeber-Einstellung
inline uint8_t stk_getClockPrescaler(void)
{
	return stk_clockPrescaler;
}

inline uint8_t stk_getClockMatch(void)
{
	return stk_clockMatch;
}

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Funktionen

inline void stk_setResponse(uint8_t status)
{
	stk_frame.dataLength = 2;
	stk_frame.data[1] = status;
}

/******************************************************************************************/

//Parameter-Anwort setzen, Lnge 3
inline void stk_setParamResponse(uint8_t param, uint8_t status)
{
	stk_frame.dataLength = 3;
	stk_frame.data[1] = status;
	stk_frame.data[2] = param;
}
