/*#######################################################################################

Copyright (C) 2013 Leo-Andres Hofmann
www.leo-andres.de

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*/

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "uart.h"

#include "general.h"
#include "settings.h"
#include "stk500.h"
#include "stk500_protocol.h"

////////////////////////////////////////////////////////////////////////////////////////////
//Makros

//Wert fr Baudraten-Register ausrechnen & runden
#define UART_IF_UBRR_VAL ( (SYSCLOCK + UART_IF_BAUDRATE * 8) / (UART_IF_BAUDRATE * 16) - 1 )

//Erreichbare Baudrate
#define UART_IF_REAL_BAUD ( SYSCLOCK / (16 * (UART_IF_UBRR_VAL + 1)) )

//Baudraten-Fehler
#define UART_IF_BAUD_ERR ( (UART_IF_REAL_BAUD * 1000) / UART_IF_BAUDRATE )

//Prfen, ob der Fehler der Baudrate unter 1 Prozent liegt
#if ( (UART_IF_BAUD_ERR < 990) || (UART_IF_BAUD_ERR > 1010) )
        #error Baudrate Error >> 1%!
#endif

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Variablen

#ifdef BOOTLOADER_PASSWORD
static const char uart_resetCode[BOOTLOADER_PASSWORD_LENGTH] PROGMEM = BOOTLOADER_PASSWORD;
#endif

volatile struct uart_status uart_status;

////////////////////////////////////////////////////////////////////////////////////////////
//Funktionen

//Initialisiert die Schnittstelle & Protokollverarbeitung
//Nach der Initialisierung ist die Schnittstelle aktiv
void uart_init(void)
{
	//Status
	uart_status.rxInFrame = FALSE;

	uart_status.txInFrame = FALSE;
	uart_status.txStart = FALSE;

#ifdef BOOTLOADER_PASSWORD
	uart_status.rxPassword = 0;
#endif

	uart_resetReceiver();

	//Baudrate
	UART_IF_UBRRH = (UART_IF_UBRR_VAL >> 8);
	UART_IF_UBRRL = (UART_IF_UBRR_VAL & 0xFF);

	//Initialisierung
	UART_IF_SET_FORMAT;
	UART_IF_SET_CONTROL;

	//Tx-Interrupt deaktivieren, keine Daten senden
	uartIf_setTxInt(FALSE);
}

/******************************************************************************************/

//Sendet den aktuellen Pufferinhalt als kompletten Frame
void uart_sendFrame(void)
{
	if( uart_status.txInFrame == FALSE )
	{
		uart_status.txInFrame = TRUE;
		uart_status.txStart = TRUE;

		uartIf_setTxInt(ON);
	}
}

/******************************************************************************************/

//Setzt die Statemachine vom Empfnger zurck
inline void uart_resetReceiver(void)
{
	uart_status.frameComplete = FALSE;
	uart_status.frameError = FALSE;
	uart_status.frameChecksumError = FALSE;
	uart_status.frameEmpty = TRUE;
	uart_status.rxInFrame = FALSE;
}

/******************************************************************************************/

//Aktiviert/deaktiviert den Interrupt zum Senden
inline void uartIf_setTxInt(uint8_t state)
{
	if( state == ON )
		UART_IF_TX_INT_ON;
	else
		UART_IF_TX_INT_OFF;
}

////////////////////////////////////////////////////////////////////////////////////////////
//Interne Funktionen


////////////////////////////////////////////////////////////////////////////////////////////
//Interrupts

//UART-Empfang vollstndig
//Dieser Interrupt wird ausgelst nachdem ein Byte empfangen wurde
ISR(UART_IF_RX_INT_VECT)
{
	unsigned char data, status;
	static unsigned char checksum = 0;
	static uint16_t bytePosition = 0;

	//Fehlercodes und Datenbyte zwischenspeichern
	status = UART_IF_GET_RX_STAT;
	data = UART_IF_DATA_REGISTER;

	//Datenempfang war fehlerhaft, Statuscode setzen
	if( status != 0x00 )
		uart_status.frameError = TRUE;

	//Puffer wird aktuell noch gesendet und jetzt berschrieben -> Fehler
	if( uart_status.txInFrame == TRUE )
		uart_status.frameError = TRUE;

	//Frame-Aufbau auswerten
	if( uart_status.rxInFrame == TRUE )
	{
		bytePosition++;

		if( bytePosition == 1 ) //Sequenznummer
		{
			stk_frame.sequence = data;

			//Frame-Puffer wird ab jetzt belegt
			uart_status.frameEmpty = FALSE;
		}
		else if( bytePosition == 2 ) //Datenlnge, MSB
		{
			stk_frame.dataLength = data;
			stk_frame.dataLength <<= 8;
		}
		else if( bytePosition == 3 ) //Datenlnge, LSB
		{
			stk_frame.dataLength |= data;

			if( stk_frame.dataLength > STK_MSG_BUFFER )
				stk_frame.dataLength = STK_MSG_BUFFER;
		}
		else if( bytePosition == 4 ) //Token
		{
			if( data != STK_FRM_TOKEN )
				uart_status.frameError = TRUE;
		}
		else if( (bytePosition > 4) && (bytePosition <= (stk_frame.dataLength + 4)) ) //Nutzdaten
		{
			stk_frame.data[bytePosition - 5] = data;
		}
		else if( bytePosition > (stk_frame.dataLength + 4) ) //Prfsumme
		{
			stk_frame.checksum = data;

			if( stk_frame.checksum != checksum )
			{
				uart_status.frameError = TRUE;
				uart_status.frameChecksumError = TRUE;
			}

			uart_status.frameComplete = TRUE;
			uart_status.rxInFrame = FALSE;
		}

		checksum ^= data;
	}
	else
	{
		//Frame-Anfang erkennen
		if( data == STK_FRM_START )
		{
			//Neuen Frame anlegen wenn der Puffer frei ist
			if( uart_status.frameEmpty == TRUE )
			{
				uart_status.rxInFrame = TRUE;
				bytePosition = 0;

#ifdef BOOTLOADER_PASSWORD
				uart_status.rxPassword = 0;
#endif
				checksum = data ^ 0;
			}
			else
				uart_status.frameError = TRUE;
		}
#ifdef BOOTLOADER_PASSWORD
		else if( stk_status.signedOn != TRUE )
		{
			//Erstes Zeichen vom Passwort empfangen: Auswertung weiterer Zeichen starten
			if( data == pgm_read_byte_near(&uart_resetCode[0]) )
			{
				uart_status.rxPassword = 1;
			}
			else if( uart_status.rxPassword > 0 )
			{
				//Zeichen weiterschalten und prfen
				if( (data == pgm_read_byte_near(&uart_resetCode[uart_status.rxPassword])) && (uart_status.rxPassword <= BOOTLOADER_PASSWORD_LENGTH))
				{
					uart_status.rxPassword++;

					//Passwort vollstndig empfangen, neustart
					if( uart_status.rxPassword == BOOTLOADER_PASSWORD_LENGTH )
						general_reboot();
				}
				else //Abbruch, Passwort unvollstndig
					uart_status.rxPassword = 0;				
			}
		}
#endif
	}
}

/******************************************************************************************/

//Senderegister leer
//Dieser Interrupt wird ausgelst, wenn das nchste Byte bertragen werden kann
ISR(UART_IF_TX_INT_VECT)
{
	unsigned char data = 0;
	static unsigned char checksum = 0;
	static uint16_t bytePosition = 0;

	if( uart_status.txInFrame == TRUE )
	{
		if( uart_status.txStart == TRUE )
		{
			bytePosition = 0;
			checksum = 0;

			uart_status.txStart = FALSE;
		}

		bytePosition++;

		if( bytePosition == 1 ) //Startblock
		{
			data = STK_FRM_START;
		}
		else if( bytePosition == 2 ) //Sequenznummer
		{
			data = stk_frame.sequence;
		}
		else if( bytePosition == 3 ) //Datenlnge, MSB
		{
			data = stk_frame.dataLength >> 8;
		}
		else if( bytePosition == 4 ) //Datenlnge, LSB
		{
			data = stk_frame.dataLength & 0x00FF;
		}
		else if( bytePosition == 5 ) //Token
		{
			data = STK_FRM_TOKEN;
		}
		else if( (bytePosition > 5) && (bytePosition <= (stk_frame.dataLength + 5)) ) //Nutzdaten
		{
			data = stk_frame.data[bytePosition - 6];
		}
		else if( bytePosition > (stk_frame.dataLength + 5) ) //Prfsumme
		{
			data = checksum;

			uart_status.txInFrame = FALSE;
		}

		checksum ^= data;

		//Daten senden
		UART_IF_DATA_REGISTER = data;		
	}
	else
		uartIf_setTxInt(OFF);
}
